#
# Reference : https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/
#
# Written by SK Park , NIRAL, UNC
# 2021-04-18

  
import dtiplayground.dmri.preprocessing as prep

import yaml
from pathlib import Path
import os
import markdown

import numpy as np
import time
import os 




# import SLICE_Check.computations as computations

class SLICE_Check(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        global logger
        logger = self.logger.write

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    @prep.measure_time
    def process(self,*args,**kwargs):  ## variables : self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        #logger(yaml.dump(inputParams))
        arte_sorted=self.slice_check(self.image,computation_dir=self.computation_dir,
                                             headskip=self.protocol['headSkipSlicePercentage'],
                                             tailskip=self.protocol['tailSkipSlicePercentage'],
                                             baseline_z_Threshold=self.protocol['correlationDeviationThresholdbaseline'],
                                             gradient_z_Threshold=self.protocol['correlationDeviationThresholdgradient'],
                                             quad_fit=self.protocol['quadFit'],
                                             subregion_check=self.protocol['bSubregionalCheck'],
                                             subregion_relaxation_factor=self.protocol['subregionalCheckRelaxationFactor']
                                             )
        logger("-------------------------------------------------------------",prep.Color.WARNING)
        logger("Abnormal gradients",prep.Color.WARNING)
        logger("-------------------------------------------------------------",prep.Color.WARNING)

        grads=self.image.getGradients()
        for a in arte_sorted:
            grad_index=a[0]
            grad_original_index=grads[a[0]]['original_index']
            vec=grads[a[0]]['gradient']
            isB0=grads[a[0]]['baseline']
            logger("For gradient {} (Org.Idx {}) , Vec {}, isB0 {}".format(grad_index,
                                                                            grad_original_index,
                                                                            vec,
                                                                            isB0))
            for i in range(len(a[1])):
                logger("\t\tSlice.idx {}, Correlation : {:.4f}".format(a[1][i]['slice'],a[1][i]['correlation']))
        gradient_indexes_to_remove=[ix[0] for ix in arte_sorted]

        ## make result and set final image to self.result and self.image (which are to be copied to the next pipeline module as on input)
        ## Excluded original indexes will be automatically deleted in the postProcess
        #logger("Excluded gradient indexes : {}".format(gradient_indexes_to_remove),prep.Color.WARNING) #gradient indexes are not original one , so need to convert
        self.result['output']['excluded_gradients_original_indexes']=self.image.convertToOriginalGradientIndex(gradient_indexes_to_remove)
        #self.result['output']['image_path']=Path(self.output_dir).joinpath('output.nrrd').__str__()
        self.result['output']['success']=True
        self.image.setSpaceDirection(target_space=self.getSourceImageInformation()['space'])
        return self.result


    @prep.measure_time
    def slice_check(self,
                    image, computation_dir, # image object containing all the information
                    headskip=0.1, 
                    tailskip=0.1, 
                    baseline_z_Threshold= 3.0, 
                    gradient_z_Threshold=3.5, 
                    quad_fit=True , 
                    subregion_check=False, ## not implemented yet
                    subregion_relaxation_factor=1.1):
        
        image_tensor=image.images.astype(float) 
        gradients=image.getGradients()
        ## Generate slice correlation informations over gradients
        gsum=[]
        begin_slice=int(np.floor(image_tensor.shape[2]*headskip))
        last_slice=int(np.floor(image_tensor.shape[2]*(1-tailskip)))
        for k in range(image_tensor.shape[3]):
            csum=[]
            for i in range(image_tensor.shape[2]):
                if i<=begin_slice or i>=last_slice: continue
                af=image_tensor[:,:,i,k].reshape([1,-1])
                bf=image_tensor[:,:,i-1,k].reshape([1,-1])
                corr=self.ncc(af,bf)
                if not np.isnan(corr): csum.append(float(corr))
                else: csum.append(0.0)
            gsum.append(csum)
        gsum=np.array(gsum) 

        ## lookup artifacts from the z-threshold criteria over gradients for each slices.
        min_bval=min(list(map(lambda x: x['b_value'],gradients)))
        max_bval=max(list(map(lambda x: x['b_value'],gradients)))
        quadfit= self.quadratic_fit_generator([min_bval,max_bval],[baseline_z_Threshold,gradient_z_Threshold])
        artifacts={}
        for slice_index in range(gsum.shape[1]):
            gradwise_vec=gsum[:,slice_index]
            avg=np.mean(gradwise_vec)
            std=np.std(gradwise_vec)
            logger("Slice {}, Mean {:.4f}, Std {:.4f}".format(slice_index+begin_slice,avg,std))
            for idx,g in enumerate(gradwise_vec):
                z_threshold=baseline_z_Threshold
                if quad_fit:
                    z_threshold=quadfit(gradients[idx]['b_value'])
                elif not gradients[idx]['baseline']:
                    z_threshold=gradient_z_Threshold
                if g < avg- z_threshold*std:
                    if idx not in artifacts: artifacts[idx]=[]
                    artifacts[idx].append({"slice":slice_index+begin_slice,
                                            "correlation":float(g),
                                            'z_threshold':float(z_threshold),
                                            'z_value':float((g-avg)/std),
                                            'b_value':float(gradients[idx]['b_value'])})

        gsum_file=Path(computation_dir).joinpath('correlation_table.yml') # row=gradient index, col = slice index
        yaml.dump(gsum.tolist(),open(gsum_file,'w'))
        artifacts_file=Path(computation_dir).joinpath("artifacts.yml")
        yaml.dump(artifacts,open(artifacts_file,'w'))

        ## recap and return the results
        arte=list(artifacts.items())
        arte_sorted=sorted(arte,key=lambda x: x[0]) ## fitst element is gradient index, second is list of slice indexes with artifacts

        return arte_sorted

    def ncc_old(self,x,y):  #normalized correlation
        x=(x-np.mean(x))/np.sqrt(np.sum((x-np.mean(x))**2))
        y=(y-np.mean(y))/np.sqrt(np.sum((y-np.mean(y))**2))
        return np.sum(x*y)

    def ncc(self,x,y): #normalized cross correlation in image (ref: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/ )
        ab=np.sum(x*y)
        a2=np.sum(x**2)
        b2=np.sum(y**2)
        if a2*b2==0.0: 
            return 1.0
        else:
            return ab/np.sqrt(a2*b2)

    def _quad_fit(self,bval, domain=[0,1000], fimage=[3.0,3.5]): #returns std multiple between fimage
        if bval > domain[1] : 
            return fimage[1]
        elif bval < domain[0] : 
            return fimage[0]
        denom=((domain[1]-domain[0])**2)
        a=0
        if denom==0: 
            a=0
        else:
            a= (fimage[1]-fimage[0])/denom
        c= fimage[0]
        return a*(bval**2)+c

    def quadratic_fit_generator(self,domain,fimage):
        def wrapper(bval):
            return self._quad_fit(bval,domain,fimage)
        return wrapper

    def makeReport(self):
        super().makeReport() 
        
        with open(os.path.abspath(self.output_dir) + '/report.md', 'a') as f:
            if len(self.result['output']['excluded_gradients_original_indexes']) == 0:
                f.write('* 0 excluded gradients\n')
            else:
                excluded_gradients = str(len(self.result['output']['excluded_gradients_original_indexes'])) + " excluded gradient(s): "
                for gradient_index in self.result['output']['excluded_gradients_original_indexes'][:-1]:
                    excluded_gradients = excluded_gradients + str(gradient_index) + ", "
                excluded_gradients += str(self.result['output']['excluded_gradients_original_indexes'][-1])
                f.write('* ' + excluded_gradients + '\n')
        self.result['report']['csv_data']['excluded_gradients'] = self.result['output']['excluded_gradients_original_indexes']
        with open(str(Path(self.output_dir).joinpath('result.yml')),'w') as f:
            yaml.dump(self.result,f)





  
import dtiplayground.dmri.preprocessing as prep
import yaml, os
from pathlib import Path

import dtiplayground.dmri.common.tools as tools 


class IDENTITY_Process(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        global logger
        logger = self.logger.write

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol
        
    @prep.measure_time
    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.output_dir, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.baseline_threshold=protocol_options['baseline_threshold']

        # << TODOS>>
        logger(yaml.dump(self.image.information))
        self.result['output']['success']=True

        return self.result




from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class DTIRegister(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, dtiregister_yml):
    QWidget.__init__(self)
    self.dtiregister_yml = dtiregister_yml
    self.method_it = 0
    self.registrationType_it = 0
    self.similarityMetric_it = 0
    self.stack = QWidget()
    self.DTIRegisterStack(protocol_template)

  def DTIRegisterStack(self, protocol_template):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "DTI_Register":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:    
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)    
    # overwrite
    self.overwrite = QCheckBox("Overwrite")    
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # method
    self.method = QComboBox()
    for ite in self.dtiregister_yml["protocol"]["method"]["candidates"]:
      self.method.addItem(ite["caption"])
      self.method.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.method.currentTextChanged.connect(self.GetMethodIt)
    self.method.setStatusTip(self.dtiregister_yml["protocol"]["method"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["method"]["caption"], self.method)
    self.method.currentTextChanged.connect(self.GetParams)
    # referenceImage
    self.referenceImage = QLineEdit()
    self.referenceImage.setAlignment(Qt.AlignRight)
    self.referenceImage.setStatusTip(self.dtiregister_yml["protocol"]["referenceImage"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["referenceImage"]["caption"], self.referenceImage)
    self.referenceImage.textChanged.connect(self.GetParams)
    # ANTsPath
    self.ANTsPath = QLineEdit()
    self.ANTsPath.setAlignment(Qt.AlignRight)
    self.ANTsPath.setStatusTip(self.dtiregister_yml["protocol"]["ANTsPath"]["description"])
    #protocol_layout.addRow(self.dtiregister_yml["protocol"]["ANTsPath"]["caption"], self.ANTsPath)
    #self.ANTsPath.textChanged.connect(self.GetParams)
    # ANTsMethod
    self.ANTsMethod = QLineEdit()
    self.ANTsMethod.setAlignment(Qt.AlignRight)
    #self.ANTsMethod.setStatusTip(self.dtiregister_yml["protocol"]["ANTsMethod"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["ANTsMethod"]["caption"], self.ANTsMethod)
    self.ANTsMethod.textChanged.connect(self.GetParams)
    # registrationType
    self.registrationType = QComboBox()
    for ite in self.dtiregister_yml["protocol"]["registrationType"]["candidates"]:
      self.registrationType.addItem(ite["caption"])
      self.registrationType.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.registrationType.currentTextChanged.connect(self.GetRegistrationTypeIt)
    self.registrationType.setStatusTip(self.dtiregister_yml["protocol"]["registrationType"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["registrationType"]["caption"], self.registrationType)
    self.registrationType.currentTextChanged.connect(self.GetParams)
    # similarityMetric
    self.similarityMetric = QComboBox()
    for ite in self.dtiregister_yml["protocol"]["similarityMetric"]["candidates"]:
      self.similarityMetric.addItem(ite["caption"])
      self.similarityMetric.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.similarityMetric.currentTextChanged.connect(self.GetSimilarityMetricIt)
    self.similarityMetric.setStatusTip(self.dtiregister_yml["protocol"]["similarityMetric"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["similarityMetric"]["caption"], self.similarityMetric)
    self.similarityMetric.currentTextChanged.connect(self.GetParams)
    # similarityParameter
    self.similarityParameter = QDoubleSpinBox()
    self.similarityParameter.setAlignment(Qt.AlignRight)
    self.similarityParameter.setSingleStep(0.1)
    self.similarityParameter.setDecimals(3)
    self.similarityParameter.setStatusTip(self.dtiregister_yml["protocol"]["similarityParameter"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["similarityParameter"]["caption"], self.similarityParameter) 
    self.similarityParameter.valueChanged.connect(self.GetParams)
    # ANTsIterations
    self.ANTsIterations = QLineEdit()
    self.ANTsIterations.setAlignment(Qt.AlignRight)
    self.ANTsIterations.setStatusTip(self.dtiregister_yml["protocol"]["ANTsIterations"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["ANTsIterations"]["caption"], self.ANTsIterations)
    self.ANTsIterations.textChanged.connect(self.GetParams)
    # gaussianSigma
    self.gaussianSigma = QDoubleSpinBox()
    self.gaussianSigma.setAlignment(Qt.AlignRight)
    self.gaussianSigma.setSingleStep(0.1)
    self.gaussianSigma.setDecimals(3)
    self.gaussianSigma.setStatusTip(self.dtiregister_yml["protocol"]["gaussianSigma"]["description"])
    protocol_layout.addRow(self.dtiregister_yml["protocol"]["gaussianSigma"]["caption"], self.gaussianSigma) 
    self.gaussianSigma.valueChanged.connect(self.GetParams)
    
    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  
  def GetMethodIt(self, text): # dti register parameter
    for it in range(len(self.dtiregister_yml["protocol"]["method"]["candidates"])):
      if self.dtiregister_yml["protocol"]["method"]["candidates"][it]["caption"] == text:
        self.method_it = it

  def GetRegistrationTypeIt(self, text): # dti register parameter
    for it in range(len(self.dtiregister_yml["protocol"]["registrationType"]["candidates"])):
      if self.dtiregister_yml["protocol"]["registrationType"]["candidates"][it]["caption"] == text:
        self.registrationType_it = it

  def GetSimilarityMetricIt(self, text): # dti register parameter
    for it in range(len(self.dtiregister_yml["protocol"]["similarityMetric"]["candidates"])):
      if self.dtiregister_yml["protocol"]["similarityMetric"]["candidates"][it]["caption"] == text:
        self.similarityMetric_it = it


  def GetParams(self):
    params = [
      'Register DTI (ANTs)',
      'DTI_Register', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'method': self.dtiregister_yml["protocol"]["method"]["candidates"][self.method_it]["value"], 
        'referenceImage': self.referenceImage.text(), 
        'ANTsPath': '',
        'ANTsMethod': self.ANTsMethod.text(),
        'registrationType': self.dtiregister_yml["protocol"]["method"]["candidates"][self.registrationType_it]["value"],
        'similarityMetric': self.dtiregister_yml["protocol"]["method"]["candidates"][self.similarityMetric_it]["value"],
        'similarityParameter': round(self.similarityParameter.value(), 3), 
        'ANTsIterations': self.ANTsIterations.text(),
        'gaussianSigma': round(self.gaussianSigma.value(), 3)
        }
      }
    ]
    self.communicate.SendParams(params)


##################################################################################
### Module : filebrowser.py
### Description : Remote(Local) File browser 
###
###
###
### Written by : scalphunter@gmail.com ,  2022/10/07
### Copyrights reserved by NIRAL
##################################################################################

import sys, traceback

### API
from flask import request, Response , send_from_directory, jsonify
import json
from . import utils
from pathlib import Path

class FileBrowserAPI:
    def __init__(self,app,**kwargs):
        self.app=app
        self.initEndpoints()

##### Endpoints

    def initEndpoints(self):
        @self.app.route('/api/v1/version',methods=['GET'])
        def _checkAppVersion():
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                res= self.checkAppVersion()
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/files',methods=['GET'])
        def _getFileList():
            sc=200
            res=None
            param_rootdir=request.args.get('root_dir',default='/',type=str)
            request_id=utils.get_request_id()
            try:
                res= self.browseDirectory(param_rootdir)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/files/get-text',methods=['GET'])
        def _getFileTextContent():
            sc=200
            res=None
            param_path=request.args.get('path',default='/',type=str)
            param_last_line=request.args.get('last_line',default=0,type=int)
            request_id=utils.get_request_id()
            try:
                res= self.getTextFileContentAsArray(param_path,param_last_line)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

        @self.app.route('/api/v1/files/get-readme',methods=['GET'])
        def _getReadMe():
            sc=200
            res=None
            request_id=utils.get_request_id()
            try:
                import dtiplayground
                path = Path(dtiplayground.__file__).resolve().parent.parent.joinpath('README.md');
                res= self.getTextFileContentAsWhole(path)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp

    def checkAppVersion(self):
        return { 'version' : "0.0.1" }

    def browseDirectory(self, rootdir):
        p = Path(rootdir).glob("*")
        parent = Path(rootdir).parent.__str__()
        fixed_paths = [
            {
                'path': parent,
                'is_dir': True,
                'name' : '..',
                'is_real': False
            }
        ]
        paths = ({'path': str(x), 'is_dir': x.is_dir(), "name": x.name, 'is_real': True } for x in p)
        paths = list(paths)
        paths.sort(key=lambda x: (not x['is_dir'], x['name']))
        return fixed_paths + list(paths) 

    def getTextFileContentAsArray(self, path, lastline):
        p = Path(path)
        with open(p,'r') as f:
            content=f.readlines()
        return content[lastline:-1]

    def getTextFileContentAsWhole(self, path):
        p = Path(path)
        with open(p,'r') as f:
            content=f.read()
        return content




##################################################################################
### Module : dmriatlasbuilder.py
### Description : Atlasbuilder api
###
###
###
### Written by : scalphunter@gmail.com ,  2022/10/08
### Copyrights reserved by NIRAL
##################################################################################

import sys, traceback

### API
from flask import request, Response , send_from_directory, jsonify
import json
from . import utils
from pathlib import Path
import yaml

from multiprocessing import Process

class DMRIAtlasbuilderAPI:
    def __init__(self,app,**kwargs):
        self.app=app
        self.initEndpoints()

##### Endpoints

    def initEndpoints(self):
        @self.app.route('/api/v1/dmriatlasbuilder/parameters',methods=['POST'])
        def _post_dmriab_params():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                req=request.get_json()
                res= self.generate_parameter_dir(req)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/dmriatlasbuilder',methods=['POST'])
        def _execute_dmriab_params():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                req=request.get_json()
                output_dir = req['output_dir']
                res= self.build_atlas(output_dir)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

    def generate_parameter_dir(self,req):
        params = {
            'output_dir' : req['output_dir'],
            'hbuild' : req['hbuild'],
            'config' : req['config'],
            'greedy' : req['greedy']
        }
        output_dir=Path(params['output_dir'])
        output_dir.mkdir(exist_ok=True,parents=False)
        common_dir = output_dir.joinpath('common')
        common_dir.mkdir(exist_ok=True)

        config_path = common_dir.joinpath('config.yml')
        hbuild_path = common_dir.joinpath('h-build.yml')
        greedy_path = common_dir.joinpath('greedy.yml')

        config = params['config']
        hbuild = params['hbuild']
        greedy = params['greedy']

        ## save config,hbuild
        yaml.safe_dump(config, open(config_path,'w'))
        yaml.safe_dump(hbuild, open(hbuild_path,'w'))
        yaml.safe_dump(greedy, open(greedy_path,'w'))

        return params

    # def build_atlas(self, configured_dir):

    #     output_dir=Path(configured_dir)
        

        
    #     config_path=output_dir.joinpath('common/config.yml')
    #     hbuild_path=output_dir.joinpath('common/h-build.yml')
    #     greedy_path=output_dir.joinpath('common/greedy.yml')
    #     ablogger = common.MultiLogger()
    #     bldr=AtlasBuilder(logger = ablogger)
    #     bldr.configure( output_dir=output_dir,
    #                     config_path=config_path,
    #                     hbuild_path=hbuild_path,
    #                     greedy_path=greedy_path)
       
    #     bldr.build()

    #     res = { 'task_id' : utils.get_uuid()}
    #     return res

    def build_atlas(self, configured_dir):

        output_dir=Path(configured_dir)
        
        def dmriatlas_proc(output_dir):
            with open(output_dir.joinpath('log.txt'),'w') as sys.stdout:
                import  dtiplayground.dmri.common as common   
                from dtiplayground.dmri.atlasbuilder import AtlasBuilder 
                logger = common.logger
                logger.setFilePointer(sys.stdout)
                config_path=output_dir.joinpath('common/config.yml')
                hbuild_path=output_dir.joinpath('common/h-build.yml')
                greedy_path=output_dir.joinpath('common/greedy.yml')
                bldr=AtlasBuilder(logger = logger)
                bldr.configure( output_dir=output_dir,
                                config_path=config_path,
                                hbuild_path=hbuild_path,
                                greedy_path=greedy_path)
           
                bldr.build()

        res = { 'task_id' : utils.get_uuid()}
        proc = Process(target= dmriatlas_proc, args=[output_dir])
        proc.start()
        proc.join()
        if proc.exitcode != 0 : raise Exception("Error during running")
        return res




##################################################################################
### Module : utils.py
### Description : Utility functions 
###
###
###
### Written by : scalphunter@gmail.com ,  2021/08/31
### Copyrighted reserved by Sentinel Holdings LC
##################################################################################


import json
from flask import request 

import hmac
import hashlib
import base64
import uuid
import os,traceback
import datetime
import random
import string 
import math

## Exception helper
def check_keys(objToValidate,keys:list):
    for k in keys:
        if k not in objToValidate: raise Exception("Key '{}' is missing".format(k))

## Utilities like loggers 

def get_secret_hash(username,client_id,client_secret):
    msg = username + client_id
    dig = hmac.new(str(client_secret).encode('utf-8'), 
        msg = str(msg).encode('utf-8'), digestmod=hashlib.sha256).digest()
    d2 = base64.b64encode(dig).decode()
    return d2

## json serialize handlers
def datetime_handler(x):
    if isinstance(x, datetime.datetime):
        return x.isoformat()
    raise TypeError("Unknown type")

## public/private
def add_public(doc,public):
    doc['is_public']=public
    return doc
## user id add 
def add_user(doc,user):
    if 'created_by' in doc:
        doc['updated_by']=user.id
        doc['updated_by_name']=user.username
    else:
        doc['created_by']=user.id
        doc['created_by_name']=user.username
        doc['updated_by']=doc['created_by']
        doc['updated_by_name']=user.username
    return doc

## doc status
def set_to_live(doc):
    doc['status']='live'

def set_to_dead(doc):
    doc['status']='dead'
    
## add access_rights
def add_access_rights(doc, user, rights = {}):
    if 'access_rights' not in doc:
        doc['access_rights']=[]
    payload = {
        'user_id': user.id,
        'username': user.username,
        'read': True,
        'write': True,
        'delete': True,
        'admin': True
    }
    payload.update(rights)
    doc['access_rights'].append(payload)
    return doc

def add_access_rights_by_user_id(doc, user_id, username, rights = {}):
    if 'access_rights' not in doc:
        doc['access_rights']=[]
    payload = {
        'user_id': user_id,
        'username': username,
        'read': True,
        'write': True,
        'delete': True,
        'admin': True
    }
    payload.update(rights)
    doc['access_rights'].append(payload)
    return doc
## timestamp
def get_timestamp():
    return datetime.datetime.now().isoformat() 

def add_timestamp(doc):
    if 'created_at' in doc:
        doc['updated_at']=get_timestamp()
    else:
        doc['created_at']=get_timestamp()
        doc['updated_at']=doc['created_at']
    return doc

## uuid

def get_uuid():
    return str(uuid.uuid4())

def add_uuid(doc):
    doc["_id"]=get_uuid()
    return doc

## request id
def get_request_id():
    return "{}_{}".format(get_timestamp(),get_uuid())

def add_request_id(doc,request_id=None):
    if request_id is None : request_id=get_request_id()
    newdoc={}
    newdoc['request_id']=request_id
    newdoc['data']=doc
    return newdoc 
    
# def add_request_id(doc,request_id=None):
#     if request_id is None : request_id=get_request_id()
#     if type(doc) == dict:
#         doc['request_id']=request_id
#         return doc
#     else:
#         newdoc={}
#         newdoc['request_id']=request_id
#         newdoc['data']=doc
#         return newdoc 

## password generator
def generate_password(length=12):
    lower=string.ascii_lowercase
    upper=string.ascii_uppercase
    num=string.digits
    symbols=string.punctuation
    allc=lower+upper+num+symbols
    temp=random.sample(allc,length)
    password="".join(temp)
    return password 

## logger

def get_user_ip():
    headers_list = request.headers.getlist("X-Forwarded-For")
    user_ip = headers_list[0] if headers_list else request.remote_addr
    return user_ip

def log(msg=""):
    username="anonymous"
    try:
        u,g=current_user    
        username=u.username 
    except Exception as e:
        pass

    s="[{} {}] <{}@{}> MSG: {} ".format(request.method,request.path,username,get_user_ip(),msg)
    return s

## dict to attributes

def dict_to_attributes(dict_obj):
    attrs=[]
    for k,v in dict_obj.items():
        attrs.append({
                'Name' : k,
                'Value' : v
            })
    return attrs 

def attributes_to_dict(attr_list):
    return [{x['Name'] : x['Value']} for x in attr_list]

## error message
def result_message(msg,status_code=200,request_id=None):
    if request_id is None: request_id=get_request_id()
    return {"msg":msg,"request_id":request_id,"status_code":status_code}

def error_message(msg,status_code,request_id=None):
    if request_id is None: request_id=get_request_id()
    return {"msg":msg,"request_id":request_id,"status_code":status_code}




#!python

import shutil
import os 
import importlib
from pathlib import Path
import argparse,yaml
from argparse import RawTextHelpFormatter
import traceback,time,copy,yaml,sys,uuid
import sys
import subprocess
import re
sys.path.append(Path(__file__).resolve().parent.parent.__str__())  ## this line is for development
import dtiplayground
import dtiplayground.dmri.common
import dtiplayground.dmri.common.module
from dtiplayground.config import INFO as info
from dtiplayground.api.server import DTIPlaygroundServer

logger=dtiplayground.dmri.common.logger.write 
color= dtiplayground.dmri.common.Color

### unit functions

def resolve_softwarepaths(spathobj, globalvars):
    if 'dtiplayground-tools' in globalvars:
        p = Path(globalvars['dtiplayground-tools']['path'])
        if p.exists():
            os.environ['DTIPLAYGROUNDTOOLS'] = p.resolve().__str__()

    if 'fsl' in globalvars:
        p = Path(globalvars['fsl']['path'])
        if p.exists():
            os.environ['FSL'] = p.resolve().__str__()

    if  'DTIPLAYGROUNDTOOLS' not in os.environ:
        os.environ['DTIPLAYGROUNDTOOLS']=os.path.expandvars("$HOME/.niral/dtiplayground-tools")
        tooldir=os.environ['DTIPLAYGROUNDTOOLS']
        if Path(tooldir).exists():
            logger("DTI Playground tools directory found at {}".format(tooldir), color.OK)
        else:
            logger("DTI Playground tools directory is set to {}".format(tooldir), color.WARNING)
            logger("DTI Playground tools directory not found, please install toolkits or specify software paths manually", color.WARNING)
    else:
        tooldir=os.environ['DTIPLAYGROUNDTOOLS']
        if Path(os.environ['DTIPLAYGROUNDTOOLS']).exists():
            
            logger("DTI playground tools are found in {}".format(tooldir),color.OK)
        else:
            logger("DTI playground tools are not found in {}".format(tooldir),color.ERROR)
    if 'FSL' not in os.environ:
        logger("FSL directory is not set in environment variable FSL",color.WARNING)
        logger("Make sure FSL directory is set in the software paths",color.WARNING)
    else:
        if Path(os.environ['FSL']).exists():
            logger("FSL directory is found at {}".format(os.environ['FSL']),color.OK)
        else:
            logger("FSL directory is set in environment variable FSL, but doesn't exist",color.WARNING)
    
    sp = spathobj['softwares']
    for k,v in sp.items():
        resolved_path = os.path.expandvars(v['path'])
        sp[k]['path'] = resolved_path
    spathobj['softwares'] = sp
    return spathobj

def initialize_logger(args):
    ## default log setting
    dtiplayground.dmri.common.logger.setLogfile(args.log)
    dtiplayground.dmri.common.logger.setTimestamp(not args.no_log_timestamp)
    dtiplayground.dmri.common.logger.setVerbosity(not args.no_verbosity)

def check_initialized(args):
    home_dir=Path(args.config_dir)
    if home_dir.exists():
        config_file=home_dir.joinpath('config.yml')
        environment_file=home_dir.joinpath('environment.yml')
        if not config_file.exists():  return False
        if not environment_file.exists(): return False
        return True
    else: 
        return False

def load_configurations(config_dir:str):
    ## reparametrization
    home_dir=Path(config_dir)
    ## Function begins
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    config=yaml.safe_load(open(config_filename,'r'))
    environment=yaml.safe_load(open(environment_filename,'r'))
    return config,environment

### decorators
def after_initialized(func): #decorator for other functions other than command_init
    def wrapper(args):
        home_dir=Path(args.config_dir)
        if home_dir.exists() and home_dir.joinpath('config.yml').exists() and home_dir.joinpath('environment.yml').exists():
            config_file=home_dir.joinpath('config.yml')
            environment_file=home_dir.joinpath('environment.yml')
            initialize_logger(args)
            return func(args)
        else: 
            # raise Exception("Not initialized, please run init command at the first use")
            logger("Configuration directory is not found or some files are missing, commencing initialization ...",dtiplayground.dmri.common.Color.WARNING)
            command_init(args)
            config_file=home_dir.joinpath('config.yml')
            environment_file=home_dir.joinpath('environment.yml')
            initialize_logger(args)
            return func(args)
    return wrapper

def log_off(func):
    def wrapper(*args,**kwargs):
        dtiplayground.dmri.common.logger.setVerbosity(False)
        res=func(*args,**kwargs)
        dtiplayground.dmri.common.logger.setVerbosity(True)
        return res 
    return wrapper

def search_fsl(lookupdirs=None):
    fslpath=None
    if 'FSLDIR' in os.environ:
        fslpath1=Path(os.environ['FSLDIR'])
        if fslpath1.exists():
            fslpath=fslpath1
    elif 'FSL' in os.environ:
        fslpath1=Path(os.environ['FSL'])
        if fslpath1.exists():
            fslpath=fslpath1
    else:
        fslpath=None

    if fslpath is not None:
        if fslpath.exists():
            logger("FSL directory found : {}".format(fslpath.resolve().__str__()), color.OK)
    else:
        logger("FSL directory not found for environment variable FSLDIR/FSL", color.WARNING)
        search_dirs=[]
        if lookupdirs is None or len(lookupdirs) < 1:
            search_dir = input("Specify the directory of installed FSL [/] :")
            if not search_dir: 
                search_dirs = ['/']
            else:
                search_dirs = [search_dir]
        else:
            logger("Looking up in the directories : {}".format(lookupdirs))
            search_dirs=lookupdirs

        for sdir in search_dirs:
            should_break=False
            search_dir = Path(sdir)
            candidates = search_dir.glob("**/bin/fsl")
            logger("Searching FSL in {}".format(search_dir.__str__()),color.PROCESS)
            for c in candidates:
                
                c = Path(c)
                if c.__str__().lower().endswith('/bin/fsl'):
                    fslpath = c.resolve().parent.parent
                    version=open(fslpath.joinpath('etc/fslversion'),'r').read().strip()
                    logger("Found : {}, Version: {}".format(fslpath.__str__(), version), color.OK)
                    confirm = input("Is this FSL you want? [Y/n]")
                    if not confirm: confirm = 'y'
                    if confirm.lower() == 'y':
                        should_break=True
                        break
            if should_break:
                break

    if fslpath.joinpath('etc/fslversion').exists():
        info = {
            "name" : "fsl",
            "version" : open(fslpath.joinpath('etc/fslversion'),'r').read().strip()
        }
        res = {
            "fsl" : {
                "path" : fslpath.resolve().__str__(),
                "info" : info
            }
        }
        return res
    else:
        return None

def search_dtiplayground_tools(lookupdirs=None):   
    target_path=None
    search_dirs=[]
    if lookupdirs is None or len(lookupdirs) < 1:
        search_dir = input("Specify the directory of installed DTI Playground tools [/] :")
        if not search_dir: 
            search_dirs = ['/']
        else:
            search_dirs = [search_dir]
    else:
        search_dirs=lookupdirs
    for sdir in search_dirs:
        should_break=False
        search_dir = Path(sdir)
        candidates = search_dir.glob("**/info.yml")
        info={}
        logger("Searching DTIPlaygroundTools in {}".format(search_dir.__str__()),color.PROCESS)
        for c in candidates:
            c = Path(c)
            info = yaml.safe_load(open(c,'r'))
            if 'name' in info:
                if info['name']=='dtiplayground-tools':
                    target_path = c.resolve().parent
                    version=info['version']
                    logger("Found : {}, Version: {}".format(target_path.__str__(), version), color.OK)
                    confirm = input("Is this DTI Playground Tools you want? [Y/n]")
                    if not confirm: confirm = 'y'
                    if confirm.lower() == 'y':
                        should_break=True
                        break
        if should_break: break

    if target_path is not None:
        res = {
            "dtiplayground-tools" : {
                "path" : target_path.resolve().__str__(),
                "info" : info
            }
        }
        return res
    else:
        return None
    
def update_software_paths(args, globalvars):
    home_dir=Path(args.config_dir)
    software_info_path=Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/common/data/software_paths.yml")
    software_paths=yaml.safe_load(open(software_info_path,'r'))
    software_paths=resolve_softwarepaths(software_paths, globalvars)
    software_filename=home_dir.joinpath('software_paths.yml')
    yaml.dump(software_paths, open(software_filename,'w'))

### command functions

def command_find_tools(args):
    ## reparametrization
    home_dir=Path(args.config_dir)

    ## Function begins
    lookup_dirs = args.lookup_dirs
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    software_filename=home_dir.joinpath('software_paths.yml')
    software_info_path=Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/common/data/software_paths.yml")
    globalvars_fn = home_dir.parent.joinpath('global_variables.yml')
    globalvars={}
    if globalvars_fn.exists():
        globalvars.update(yaml.safe_load(open(globalvars_fn,'r')))

    ## load default software paths

    software_paths=yaml.safe_load(open(software_info_path,'r'))
    result = False
    ## search FSL
    fsl_info = search_fsl(lookupdirs=lookup_dirs)
    if fsl_info is not None:
        globalvars.update(fsl_info) 
        logger("Selected FSL directory is : {}".format(fsl_info['fsl']['path']),color.INFO)
    else:
        logger("FSL was not found",color.ERROR)
    ## search DTI Playground Tools
    
    dpt_info = search_dtiplayground_tools(lookupdirs=lookup_dirs)
    if dpt_info is not None:
        globalvars.update(dpt_info)
        logger("Selected DTI Playground tools directory is : {}".format(dpt_info['dtiplayground-tools']['path']),color.INFO)
    else:
        logger("DTI Playground tools were not found",color.ERROR)
    ## update software paths
    logger("Updating global_variables.yml", color.PROCESS)
    yaml.dump(globalvars, open(globalvars_fn,'w'))
    logger("Updating software paths",color.PROCESS)
    update_software_paths(args, globalvars)
    logger("Updates Completed",color.OK)
    result= True
    return result

def command_install_tools(args):
    ## reparametrization
    home_dir=Path(args.config_dir).resolve()
    home_dir.parent.mkdir(parents=True,exist_ok=True)
    # globalvars_fn = Path(os.path.expandvars("$HOME/.niral-dti/global_variables.yml"))
    globalvars_fn = home_dir.parent.joinpath('global_variables.yml')
    Path(globalvars_fn).parent.mkdir(parents=True,exist_ok=True)
    globalvars = {}
    if globalvars_fn.exists():
        globalvars = yaml.safe_load(open(globalvars_fn,'r'))

    rootdir = Path(os.path.expandvars(args.output_dir)).resolve()
    rootdir.mkdir(parents=True,exist_ok=True)
    tempdir = rootdir.joinpath('temp-dtiplayground-tools')
    outputdir = rootdir.joinpath('dtiplayground-tools')
    fsldir = rootdir.joinpath('FSL')
    clean_install = args.clean_install
    leave_files_after_installation = args.no_remove
    no_fsl = args.no_fsl
    install_only = args.install_only
    build=args.build
    ### Clean install preparations
    if clean_install:
        logger("Removing temporary, existing packages (dtiplayground-tools only) directories ...")
        if tempdir.exists():
            shutil.rmtree(tempdir)
        if outputdir.exists():
            shutil.rmtree(outputdir)
        logger("Removing files finished")

    ### FSL
    if not no_fsl and not fsldir.exists():
        fslinstaller_fn=Path(__file__).resolve().parent.joinpath('fslinstaller.py')
        command=['/usr/bin/python', fslinstaller_fn ,'-d',fsldir.resolve().__str__()] ### python2 is needed
        subprocess.run(command)
        info = {
            'name' : 'fsl',
            'version' : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
        }
        globalvars['fsl']={
            'path' : fsldir.resolve().__str__(),
            'info' : info
        }
        yaml.dump(info, open(fsldir.joinpath('info.yml'),'w'))
    else:
        if fsldir.exists():
            info = {
                'name': 'fsl',
                'version' : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
            }
            yaml.dump(info, open(fsldir.joinpath('info.yml'),'w'))
            globalvars['fsl']={
                'path' : fsldir.resolve().__str__(),
                'info' : info
            }            

    ### DTI playground tools (build with docker, centos7 for now)
    tempdir.mkdir(parents=True,exist_ok=True)
    os.chdir(tempdir)
    srcdir = tempdir.joinpath('dtiplaygroundtools')
    info={}
    if not outputdir.exists():
        if build:
            if not srcdir.exists():
                logger("Fetching source code ...")
                fetch_dtiplaygroundtools = ["git","clone","https://github.com/niraluser/dtiplaygroundtools.git"]
                subprocess.run(fetch_dtiplaygroundtools)
                logger("Source codes downloaded")
            
            os.chdir(srcdir.joinpath('dockerfiles'))
            build_command = ['./build.sh']
            logger("Building software packages ... ")
            subprocess.run(build_command)
            tar_filename = srcdir.joinpath('dist/dtiplayground-tools.tar.gz')
            untar_command = ['tar','xvfz',tar_filename.resolve().__str__(), '-C', rootdir.resolve().__str__()]
            logger("Installing softwares to the output directory {}".format(outputdir.resolve().__str__()))
            subprocess.run(untar_command)
            ### read version info
            info_fn = outputdir.joinpath('info.yml')
            if info_fn.exists():
                info = yaml.safe_load(open(info_fn,'r'))
        else: # prebuilt package (default)
            tar_filename = outputdir.joinpath('dtiplayground-tools.tar.gz')
            if not tar_filename.exists():
                remote_package = "https://github.com/NIRALUser/DTIPlaygroundTools/releases/download/v0.0.1/dtiplayground-tools.tar.gz"
                fetch_package = ['wget',remote_package,'-P',outputdir.resolve().__str__()]
                subprocess.run(fetch_package)
            untar_command = ['tar','xvfz',tar_filename.resolve().__str__(), '-C', rootdir.resolve().__str__()]
            logger("Installing softwares to the output directory {}".format(outputdir.resolve().__str__()))
            subprocess.run(untar_command)
            info_fn = outputdir.joinpath('info.yml')
            if info_fn.exists():
                info = yaml.safe_load(open(info_fn,'r'))
            logger("Removing temporary files ...")
            tar_filename.unlink()

    else:
        info = yaml.safe_load(open(outputdir.joinpath('info.yml'),'r'))
    globalvars['dtiplayground-tools']={
        'path' : outputdir.resolve().__str__(),
        'info' : info
    }

    if not leave_files_after_installation:
        logger("Removing temporary files ...")
        shutil.rmtree(tempdir)
        logger("Removing temporary files finished")

    logger("Saving tool information to {}".format(globalvars_fn.__str__()))
    yaml.dump(globalvars, open(globalvars_fn,'w'))
    if home_dir.joinpath('software_paths').exists() and not install_only:
        update_software_paths(args, globalvars)
    logger("Installation completed")
    return True


def command_init(args):
    ## reparametrization
    home_dir=Path(args.config_dir)

    ## Function begins
    home_dir.mkdir(parents=True,exist_ok=True)
    tools_dir = None
    if args.tools_dir is not None:
        tools_dir = Path(args.tools_dir)

    user_module_dir=home_dir.joinpath('modules').absolute()
    user_module_dir.mkdir(parents=True,exist_ok=True)

    globalvars_fn=home_dir.joinpath('global_variables.yml')
    globalvars = {}

    if globalvars_fn.exists():
        globalvars = yaml.safe_load(open(globalvars_fn,'r'))

    if tools_dir is not None:
        if tools_dir.exists():
            dptdir = tools_dir.joinpath('dtiplayground-tools')
            fsldir = tools_dir.joinpath('FSL')
            globalvars.update({
                "dtiplayground-tools" :{
                    "path" : dptdir.resolve().__str__(),
                    "info" : yaml.safe_load(open(dptdir.joinpath('info.yml'),'r'))
                },
                "fsl" : {
                    "path" : fsldir.resolve().__str__(),
                    "info" : {
                        'name' : 'fsl',
                        "verson" : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
                    }
                }
            })
            yaml.dump(globalvars, open(globalvars_fn,'w'))
        else:
            logger("Couldn't find tool directory {}".format(str(tools_dir)),color.ERROR)
            exit(1)

    initialize_logger(args)
    logger("Initialized",color.OK)
    return True

def command_server(args):
    config = {
        "config_dir": args.config_dir,
        "host" : args.host,
        "port" : args.port,
        "static_page_dir" : args.directory,
        "browser" : args.browser
    }
    config_dir = Path(config['config_dir'])

    ## spa installation
    logger("Installing UI",color.PROCESS)
    static_dir = config_dir.joinpath('static')
    import dtiplayground.api.static.spa as s 
    from zipfile import ZipFile
    spa_archive_fn=Path(s.__file__).resolve().parent.joinpath('spa.zip')
    with ZipFile(spa_archive_fn,'r') as z:
        z.extractall(static_dir)
    logger("Done",color.OK)
    logger("API Server initiating ... ",color.PROCESS)
    
    ## app prep
    app = DTIPlaygroundServer()
    app.configure(host=config['host'],
                  port=config['port'],
                  static_folder=config['static_page_dir'],
                  static_url_path='/')

    if config['browser']: 
        import webbrowser
        from threading import Timer
        def open_browser():
            webbrowser.open('http://localhost:6543')
        Timer(1, open_browser).start();
    logger("API Server initiated, running",color.OK)
    app.serve()

    return True

### Arguments 

def get_args():
    current_dir=Path(__file__).resolve().parent
    # info=yaml.safe_load(open(current_dir.parent.joinpath('dtiplayground/info.json'),'r'))
    version=info['dtiplayground']['version']
    logger("VERSION : {}".format(str(version)))
    config_dir=Path(os.environ.get('HOME')).joinpath('.niral-dti')
    # ## read template
    module_help_str=None
    if config_dir.exists() and config_dir.joinpath('config.yml').exists() and config_dir.joinpath('environment.yml').exists():
        config,environment = load_configurations(str(config_dir))
        template_path=config_dir.joinpath(config['protocol_template_path'])
        template=yaml.safe_load(open(template_path,'r'))
        available_modules=template['options']['execution']['pipeline']['candidates']
        available_modules_list=["{}".format(x['value'])  for x in available_modules if x['description']!="Not implemented"]
        module_help_str="Avaliable Modules := \n" + " , ".join(available_modules_list)
        # print(module_help_str)
    uid, ts = dtiplayground.dmri.common.get_uuid(), dtiplayground.dmri.common.get_timestamp()

    ### Argument parsers

    parser=argparse.ArgumentParser(prog="dpg",
                                   formatter_class=RawTextHelpFormatter,
                                   description="The dtiplayground is an integrated DWI processing platform",
                                   epilog="Written by SK Park (sangkyoon_park@med.unc.edu) , Johanna Dubos (johannadubos32@gmail.com) , Neuro Image Research and Analysis Laboratories, University of North Carolina @ Chapel Hill , United States, 2021")
    subparsers=parser.add_subparsers(help="Commands")
    
    ## init command
    parser_init=subparsers.add_parser('init',help='Initialize configurations')
    parser_init.set_defaults(func=command_init)

    ## software find command
    parser_find_tools=subparsers.add_parser('find-tools',help='Search and update software paths')
    parser_find_tools.add_argument('-l','--lookup-dirs', help='Lookup dir list', type=str, nargs='*',required=False)
    parser_find_tools.set_defaults(func=command_find_tools)

    ## software install command
    parser_install_tools=subparsers.add_parser('install-tools',help='Install DTIPlaygroundTools')
    parser_install_tools.add_argument('-o','--output-dir', help="output directory", default="$HOME/.niral-dti")
    parser_install_tools.add_argument('-c','--clean-install', help="Remove existing files", default=False, action="store_true")
    parser_install_tools.add_argument('-b','--build', help="Build DTIPlaygroundTools", default=False, action="store_true")
    parser_install_tools.add_argument('--no-remove', help="Do not remove source and build files after installation", default=False,action="store_true")
    parser_install_tools.add_argument('--no-fsl', help="Do not install FSL", default=False, action="store_true")
    parser_install_tools.add_argument('--install-only', help="Do not update current software paths", default=False, action="store_true")
    parser_install_tools.set_defaults(func=command_install_tools)

    ## DPG Server
    parser_server=subparsers.add_parser('serve',help='DTIPlayground Server')
    parser_server.add_argument('--host', help="Host", default="127.0.0.1")
    parser_server.add_argument('-p','--port', help="Port", default=6543)
    parser_server.add_argument('-d','--directory', help="Static Page Path", default=str(config_dir.joinpath('static/spa')))
    parser_server.add_argument('--browser', help="Launch browser at start up", default=False, action="store_true")
    parser_server.set_defaults(func=command_server)

    ## log related
    parser.add_argument('--config-dir',help='Configuration directory',default=str(config_dir))
    parser.add_argument('--log',help='log file',default=str(config_dir.joinpath('log.txt')))
    parser.add_argument('--execution-id',help='execution id',default=uid,type=str)
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")
    parser.add_argument('-v','--version', help="Show version", default=False,action="store_true")
    parser.add_argument('--tools-dir', help="Initialize with specific tool directory", default=None)

    ## if no parameter is furnished, exit with printing help
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args=parser.parse_args()
    if args.version:
        sys.exit(1)

    return args 

## threading environment
args=get_args()
if hasattr(args,'num_threads'):
    os.environ['OMP_NUM_THREADS']=str(args.num_threads) ## this should go before loading any dipy function. 
    os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(args.num_threads) ## for ANTS threading

if __name__=='__main__':
    try:
        dtiplayground.dmri.common.logger.setTimestamp(True)
        result=args.func(args)
        exit(0)
    except Exception as e:
        dtiplayground.dmri.common.logger.setVerbosity(True)
        msg=traceback.format_exc()
        logger(msg,color.ERROR)
        exit(-1)
    finally:
        pass





import sys,os
from setuptools import setup, find_packages
import json
from os.path import join as pjoin, dirname, exists
from glob import glob
from dtiplayground.config import INFO as info

using_setuptools = 'setuptools' in sys.modules
extra_setuptools_args = {}

if using_setuptools:
    # Set setuptools extra arguments
    extra_setuptools_args = dict(
        tests_require=[],
        zip_safe=False,
        python_requires=">= 3.8",
        )

    
setup(
    name='dtiplayground',
    version=info['dtiplayground']['version'],
    python_requires=">=3.8",
    license='MIT',
    author="SK Park, NIRAL, University of North Carolina @ Chapel Hill",
    author_email='scalphunter@gmail.com',
    packages=find_packages('.'),
    package_dir={'':'.'},
    package_data = {
    '': ['*.yml','*.yaml','*.json','*.xml','*.cnf','*.md','*.zip']
    },
    scripts=glob(pjoin('bin', '*')),
    url='https://github.com/niraluser/dtiplayground',
    keywords=['dtiplayground','dmriprep','dmriatlas','dmriautotract','dmrifiberprofile','nrrd','nifti','dwi','dti','qc','quality control'],
    install_requires=[
        'flask',
        'flask_cors',
        'cmake>=3.24.1',
        'pynrrd>=0.4.2',
        'dipy>=1.4.0',
        # 'fury==0.7.0',
        # 'fury',
        'pyyaml>=5.3.1',
        'nibabel>=3.2.1',
        'tensorflow==2.8.0',
        'antspyx==0.3.2', #0.3.3 and above has some build issue
        'antspynet==0.1.8',
        'pandas>=1.4.3',
        'pyqt5',
        # 'simpleitk==2.1.1',
        'simpleitk>=2.1.1',
        'reportlab>=3.6.6',
        'pypdf2>=1.26.0',
        'markdown>=3.3.6',
        'xhtml2pdf>=0.2.7',
        'fury>=0.7.0',
        'sphinx',
        'sphinx_rtd_theme'
       ],

 )




import os 
from pathlib import Path
import traceback
import sys

from flask import Flask
from flask import send_from_directory
from flask_cors import CORS


import dtiplayground
import dtiplayground.dmri.common as common
from dtiplayground.config import INFO as info

logger=dtiplayground.dmri.common.logger.write 
color= dtiplayground.dmri.common.Color

### APIs
from dtiplayground.api.filebrowser import FileBrowserAPI 
from dtiplayground.api.dmriatlasbuilder import DMRIAtlasbuilderAPI
from dtiplayground.api.dmriprep import DMRIPrepAPI

class DTIPlaygroundServer(object):
    def __init__(self,*args,**kwargs):
        self.host = '127.0.0.1'
        self.port = '6543'
        self.static_folder = None
        self.static_url_path='/'
        self.app = None

    def configure(self, host, port, static_folder = None , static_url_path='/'):

        self.host=host
        self.port=port
        self.static_folder=static_folder
        self.static_url_path = static_url_path

        self.app=Flask(__name__, static_folder=self.static_folder, static_url_path=self.static_url_path)
        CORS(self.app)

        self.app.config['SUBMODULES']={}
        self.app.config['SUBMODULES']['FileBrowserAPI']=FileBrowserAPI(self.app)
        self.app.config['SUBMODULES']['DMRIAtlasbuilderAPI']=DMRIAtlasbuilderAPI(self.app)
        self.app.config['SUBMODULES']['DMRIPrepAPI']=DMRIPrepAPI(self.app)

        @self.app.route('/', defaults={'path': ''})
        @self.app.route('/<path>') ## for routers
        def catch_all(path):
            return self.app.send_static_file("index.html")


    def get_app(self): ## for UWSGI
        return self.app

    def serve(self):
        if not self.app: raise Exception('Flask app is not configured')
        self.app.run(host=self.host, port=self.port,debug=True)




from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class BrainTractography(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, braintractography_yml):
    QWidget.__init__(self)
    self.braintractography_yml = braintractography_yml
    self.whiteMatterMaskThreshold_it = 0
    self.method_it = 0
    self.stack = QWidget()
    self.BrainTractographyStack(protocol_template)

  def BrainTractographyStack(self, protocol_template):

    ## Module
    self.tab_name = QLabel()
    
    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "BRAIN_Tractography":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")

    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_sublayout = QFormLayout()
    protocol_layout = QVBoxLayout()
    protocol_layout.addLayout(protocol_sublayout)
    protocol_groupbox.setLayout(protocol_layout)
    # whiteMatterMaskThreshold
    self.whiteMatterMaskThreshold = QComboBox()
    for ite in self.braintractography_yml["protocol"]["whiteMatterMaskThreshold"]["candidates"]:
      self.whiteMatterMaskThreshold.addItem(ite["caption"])
      self.whiteMatterMaskThreshold.setItemData(self.whiteMatterMaskThreshold.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.whiteMatterMaskThreshold.currentTextChanged.connect(self.GetWhiteMatterMaskThresholdIt)
    self.whiteMatterMaskThreshold.setStatusTip(self.braintractography_yml["protocol"]["whiteMatterMaskThreshold"]["description"])
    protocol_sublayout.addRow(self.braintractography_yml["protocol"]["whiteMatterMaskThreshold"]["caption"], self.whiteMatterMaskThreshold)
    self.whiteMatterMaskThreshold.currentTextChanged.connect(self.GetParams)
    self.whiteMatterMaskThreshold.currentTextChanged.connect(self.DisplayWhiteMatterMaskThresholds)
    # thresholdLow
    self.thresholdLow = QDoubleSpinBox()
    self.thresholdLow.setAlignment(Qt.AlignRight)
    self.thresholdLow.setSingleStep(0.1)
    self.thresholdLow.setDecimals(3)
    self.thresholdLow.setStatusTip(self.braintractography_yml["protocol"]["thresholdLow"]["description"])
    self.thresholdLow_label = QLabel(self.braintractography_yml["protocol"]["thresholdLow"]["caption"])
    protocol_sublayout.addRow(self.thresholdLow_label, self.thresholdLow)
    self.thresholdLow.valueChanged.connect(self.GetParams)
    # thresholdUp
    self.thresholdUp = QDoubleSpinBox()
    self.thresholdUp.setAlignment(Qt.AlignRight)
    self.thresholdUp.setSingleStep(0.1)
    self.thresholdUp.setDecimals(3)
    self.thresholdUp.setStatusTip(self.braintractography_yml["protocol"]["thresholdUp"]["description"])
    self.thresholdUp_label = QLabel(self.braintractography_yml["protocol"]["thresholdUp"]["caption"])
    protocol_sublayout.addRow(self.thresholdUp_label, self.thresholdUp)
    self.thresholdUp.valueChanged.connect(self.GetParams)
    # method
    self.method = QComboBox()
    for ite in self.braintractography_yml["protocol"]["method"]["candidates"]:
      self.method.addItem(ite["caption"])
      self.method.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.method.currentTextChanged.connect(self.GetMethodIt)
    self.method.setStatusTip(self.braintractography_yml["protocol"]["method"]["description"])
    protocol_sublayout.addRow(self.braintractography_yml["protocol"]["method"]["caption"], self.method)
    self.method.currentTextChanged.connect(self.GetParams)
    # shOrder
    self.shOrder = QSpinBox()
    self.shOrder.setAlignment(Qt.AlignRight)
    self.shOrder.setSingleStep(2)
    self.shOrder.setStatusTip(self.braintractography_yml["protocol"]["shOrder"]["description"])
    self.shOrder_label = QLabel(self.braintractography_yml["protocol"]["shOrder"]["caption"])
    protocol_sublayout.addRow(self.shOrder_label, self.shOrder)
    self.shOrder.valueChanged.connect(self.GetParams)
    # relativePeakThreshold
    self.relativePeakThreshold = QDoubleSpinBox()
    self.relativePeakThreshold.setAlignment(Qt.AlignRight)
    self.relativePeakThreshold.setSingleStep(0.1)
    self.relativePeakThreshold.setDecimals(3)
    self.relativePeakThreshold.setStatusTip(self.braintractography_yml["protocol"]["relativePeakThreshold"]["description"])
    protocol_sublayout.addRow(self.braintractography_yml["protocol"]["relativePeakThreshold"]["caption"], self.relativePeakThreshold)
    self.relativePeakThreshold.valueChanged.connect(self.GetParams)
    # minPeakSeparationAngle
    self.minPeakSeparationAngle = QDoubleSpinBox()
    self.minPeakSeparationAngle.setAlignment(Qt.AlignRight)
    self.minPeakSeparationAngle.setSingleStep(1)
    self.minPeakSeparationAngle.setDecimals(3)
    self.minPeakSeparationAngle.setMaximum(90)
    self.minPeakSeparationAngle.setStatusTip(self.braintractography_yml["protocol"]["minPeakSeparationAngle"]["description"])
    protocol_sublayout.addRow(self.braintractography_yml["protocol"]["minPeakSeparationAngle"]["caption"], self.minPeakSeparationAngle)
    self.minPeakSeparationAngle.valueChanged.connect(self.GetParams)
    # stoppingCriterionThreshold
    self.stoppingCriterionThreshold = QDoubleSpinBox()
    self.stoppingCriterionThreshold.setAlignment(Qt.AlignRight)
    self.stoppingCriterionThreshold.setSingleStep(0.1)
    self.stoppingCriterionThreshold.setDecimals(3)
    self.stoppingCriterionThreshold.setStatusTip(self.braintractography_yml["protocol"]["stoppingCriterionThreshold"]["description"])
    protocol_sublayout.addRow(self.braintractography_yml["protocol"]["stoppingCriterionThreshold"]["caption"], self.stoppingCriterionThreshold)
    self.stoppingCriterionThreshold.valueChanged.connect(self.GetParams)
    # vtk42
    self.vtk42_true = QRadioButton("True")
    self.vtk42_false = QRadioButton("False")
    vtk42_layout = QHBoxLayout()
    vtk42_layout.addWidget(self.vtk42_true)
    vtk42_layout.addWidget(self.vtk42_false)
    self.vtk42_buttongroup = QButtonGroup()
    self.vtk42_buttongroup.addButton(self.vtk42_true)
    self.vtk42_buttongroup.addButton(self.vtk42_false)
    self.vtk42_true.toggled.connect(self.GetParams)
    self.vtk42_true.setStatusTip(self.braintractography_yml["protocol"]["vtk42"]["description"])
    self.vtk42_false.setStatusTip(self.braintractography_yml["protocol"]["vtk42"]["description"])
    protocol_sublayout.addRow(self.braintractography_yml["protocol"]["vtk42"]["caption"], vtk42_layout)
    # removeShortTracts
    self.removeShortTracts = QCheckBox(self.braintractography_yml["protocol"]["removeShortTracts"]["caption"])
    self.removeShortTracts.stateChanged.connect(self.GetParams)
    self.removeShortTracts.stateChanged.connect(self.EnableShortTractsThreshold)
    self.shortTractsThreshold = QDoubleSpinBox()
    self.shortTractsThreshold.valueChanged.connect(self.GetParams)
    self.shortTractsThreshold.setEnabled(False)
    self.shortTractsThreshold.setMaximum(500)
    shortTracts_layout = QHBoxLayout()
    shortTracts_layout.addWidget(self.removeShortTracts)
    shortTracts_layout.addWidget(QLabel("Threshold:"))
    shortTracts_layout.addWidget(self.shortTractsThreshold)
    protocol_layout.addLayout(shortTracts_layout)
    # removeLongTracts
    self.removeLongTracts = QCheckBox(self.braintractography_yml["protocol"]["removeLongTracts"]["caption"])
    self.removeLongTracts.stateChanged.connect(self.GetParams)
    self.removeLongTracts.stateChanged.connect(self.EnableLongTractsThreshold)
    self.longTractsThreshold = QDoubleSpinBox()
    self.longTractsThreshold.valueChanged.connect(self.GetParams)
    self.longTractsThreshold.setEnabled(False)
    self.longTractsThreshold.setMaximum(500)
    longTracts_layout = QHBoxLayout()
    longTracts_layout.addWidget(self.removeLongTracts)
    longTracts_layout.addWidget(QLabel("Threshold:"))
    longTracts_layout.addWidget(self.longTractsThreshold)
    protocol_layout.addLayout(longTracts_layout)

    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def GetWhiteMatterMaskThresholdIt(self, text): # get "whiteMatterMaskThreshold" parameter
    self.whiteMatterMaskThreshold_it = 0
    for it in range(len(self.braintractography_yml["protocol"]["whiteMatterMaskThreshold"]["candidates"])):
      if self.braintractography_yml["protocol"]["whiteMatterMaskThreshold"]["candidates"][it]["caption"] == text:
        self.whiteMatterMaskThreshold_it = it

  def GetMethodIt(self, text): # get "method" parameter
    for it in range(len(self.braintractography_yml["protocol"]["method"]["candidates"])):
      if self.braintractography_yml["protocol"]["method"]["candidates"][it]["caption"] == text:
        self.method_it = it

  def DisplayWhiteMatterMaskThresholds(self):
    if self.whiteMatterMaskThreshold.currentText() != "Manual Threshold":
      self.thresholdLow_label.setHidden(True)
      self.thresholdLow.setHidden(True)
      self.thresholdUp_label.setHidden(True)
      self.thresholdUp.setHidden(True)
    else:
      self.thresholdLow_label.setHidden(False)
      self.thresholdLow.setHidden(False)
      self.thresholdUp_label.setHidden(False)
      self.thresholdUp.setHidden(False)

  def DisplayShOrder(self):
    if self.method.currentText() not in ["CSA model", "OPDT model"]:
      self.shOrder.setHidden(True)
      self.shOrder_label.setHidden(True)
    else:
      self.shOrder.setHidden(False)
      self.shOrder_label.setHidden(False)

  def EnableShortTractsThreshold(self):
    if self.removeShortTracts.isChecked():
      self.shortTractsThreshold.setEnabled(True)
    else:
      self.shortTractsThreshold.setEnabled(False)

  def EnableLongTractsThreshold(self):
    if self.removeLongTracts.isChecked():
      self.longTractsThreshold.setEnabled(True)
    else:
      self.longTractsThreshold.setEnabled(False)

  def GetParams(self):
    params = [
      'Brain Tractography',
      'BRAIN_Tractography', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        }, 
      'protocol': {
        'whiteMatterMaskThreshold': self.braintractography_yml["protocol"]["whiteMatterMaskThreshold"]["candidates"][self.whiteMatterMaskThreshold_it]["value"],
        'thresholdLow': self.thresholdLow.value(),
        'thresholdUp': self.thresholdUp.value(),
        'method': self.braintractography_yml["protocol"]["method"]["candidates"][self.method_it]["value"],
        'shOrder': self.shOrder.value(),
        'relativePeakThreshold': self.relativePeakThreshold.value(),
        'minPeakSeparationAngle': self.minPeakSeparationAngle.value(),
        'stoppingCriterionThreshold': self.stoppingCriterionThreshold.value(),
        'vtk42': self.vtk42_true.isChecked(),
        'removeShortTracts': self.removeShortTracts.isChecked(),
        'shortTractsThreshold': self.shortTractsThreshold.value(),
        'removeLongTracts': self.removeLongTracts.isChecked(),
        'longTractsThreshold': self.longTractsThreshold.value()
        }
      }
    ]
    self.communicate.SendParams(params)